/*
 * Copyright (c) 2018 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

#include <zephyr/types.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/kernel.h>
#include <zephyr/drivers/gpio.h>
#include <soc.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>

#include <bluetooth/services/lbs.h>

#include <zephyr/settings/settings.h>

#include <dk_buttons_and_leds.h>

#include <assert.h>
#include <string.h>
#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>
#include <zephyr/device.h>
#include<zephyr/drivers/flash.h>
#include <zephyr/storage/flash_map.h>
#include <zephyr/drivers/flash/nrf_qspi_nor.h>
#include <zephyr/pm/device.h>

#include "xip_code.h"

#define STORAGE_PARTITION_LABEL	storage_partition
#define STORAGE_PARTITION_ID	FIXED_PARTITION_ID(STORAGE_PARTITION_LABEL)

#define FLASH_NODE DT_INST(0, nordic_qspi_nor)
#define UART_NODE DT_CHOSEN(zephyr_console)

#define DEVICE_NAME             CONFIG_BT_DEVICE_NAME
#define DEVICE_NAME_LEN         (sizeof(DEVICE_NAME) - 1)


#define RUN_STATUS_LED          DK_LED1
#define CON_STATUS_LED          DK_LED2
#define RUN_LED_BLINK_INTERVAL  1000

#define USER_LED                DK_LED3

#define USER_BUTTON             DK_BTN1_MSK

#define STACKSIZE 1024
#define THREAD0_PRIORITY 7
#define THREAD1_PRIORITY 7


static void connected(struct bt_conn *conn, uint8_t err)
{
	if (err) {
		printk("Connection failed (err %u)\n", err);
		return;
	}

	printk("Connected\n");

	dk_set_led_on(CON_STATUS_LED);
}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	printk("Disconnected (reason %u)\n", reason);

	dk_set_led_off(CON_STATUS_LED);
}

#ifdef CONFIG_BT_LBS_SECURITY_ENABLED
static void security_changed(struct bt_conn *conn, bt_security_t level,
			     enum bt_security_err err)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (!err) {
		printk("Security changed: %s level %u\n", addr, level);
	} else {
		printk("Security failed: %s level %u err %d\n", addr, level,
			err);
	}
}
#endif

BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected        = connected,
	.disconnected     = disconnected,
};



void test_write_ext_flash(){
  static uint8_t write_buf[4096];
  static uint8_t read_buf[4096];
  static uint8_t zero_buf[4096];

	for (int i = 0; i < sizeof(write_buf); i++) {
		write_buf[i] = i;
		zero_buf[i] = 0xff;
	}
#define TEST_PARTITION test_area

  const struct device *flash_device = FIXED_PARTITION_DEVICE(TEST_PARTITION);
  const int flash_offset = FIXED_PARTITION_OFFSET(TEST_PARTITION);
  const size_t flash_size = FIXED_PARTITION_SIZE(TEST_PARTITION);

  size_t offset = 0;

	if (!device_is_ready(flash_device)) {
		printk("Flash device %s is not ready\n", flash_device->name);
		return;
	}
  while(1){
    int rc;


    if (offset >= flash_size) {
      return;
    }

    /* rc = flash_erase(flash_device, flash_offset + offset, sizeof(write_buf)); */
    /* __ASSERT(rc == 0, "ERASE failed at offset %zu", offset); */
    /*  */
    /* rc = flash_read(flash_device, flash_offset + offset, read_buf, sizeof(read_buf)); */
    /*  */
    /* __ASSERT(rc == 0, "READ ERASE failed at offset %zu", offset); */
    /* __ASSERT(memcmp(read_buf, zero_buf, sizeof(read_buf)) == 0, "VERIFY ERASE failed at offset %zu", offset); */

    rc = flash_write(flash_device, flash_offset + offset, write_buf, sizeof(write_buf));
    __ASSERT(rc == 0, "WRITE failed at offset %zu", offset);
    rc = flash_read(flash_device, flash_offset + offset, read_buf, sizeof(read_buf));
    __ASSERT(memcmp(read_buf, write_buf, sizeof(write_buf)) == 0, "READ failed at offset %zu", offset);
    __ASSERT(memcmp(read_buf, write_buf, sizeof(write_buf)) == 0, "VERIFY failed at offset %zu", offset);
    offset += sizeof(write_buf);
  }
}

static uint8_t xip_enabled = true;

void thread0(void){
  while(1){
    k_sleep(K_MSEC(100));
    if(xip_enabled){
      printk("THREAD0: ");
      xip_test_print();
    }
  }
}

void thread1(void){
  while(1){
    k_sleep(K_MSEC(100));
    printk("THREAD1: Hello from internal flash\n");
  }
}
K_THREAD_DEFINE(thread0_id, STACKSIZE, thread0, NULL, NULL, NULL,
		THREAD0_PRIORITY, 0, 0);

/* K_THREAD_DEFINE(thread1_id, STACKSIZE, thread1, NULL, NULL, NULL, */
/* 		THREAD1_PRIORITY, 0, 0); */


int main(void)
{
	int blink_status = 0;
	int err;

	printk("Starting Bluetooth Peripheral LBS example\n");

	err = dk_leds_init();
	if (err) {
		printk("LEDs init failed (err %d)\n", err);
		return 0;
	}


	err = bt_enable(NULL);
	if (err) {
		printk("Bluetooth init failed (err %d)\n", err);
		return 0;
	}

	printk("Bluetooth initialized\n");

	if (IS_ENABLED(CONFIG_SETTINGS)) {
		settings_load();
	}

	const struct device *flash_dev = DEVICE_DT_GET(FLASH_NODE);
	const struct device *uart_dev = DEVICE_DT_GET(UART_NODE);

  printk("XIP test sample\n");
  k_sleep(K_SECONDS(1));

  for (;;) {
    printk("Start test loop\n");

#if CONFIG_NORDIC_QSPI_NOR_XIP
    nrf_qspi_nor_xip_enable(flash_dev, true);
    for(int i = 0; i < 10; i++){
      xip_test_write_ext_flash();
      k_sleep(K_MSEC(1));
      printk("XIP external flash write %d\n",i);
    }
#endif

    for(int i = 0; i < 10; i++){
      test_write_ext_flash();
      k_sleep(K_MSEC(1));
      printk("External flash write %d\n",i);
    }

    printk("Disable test loop and sleep for 10 sec\n");
    pm_device_action_run(uart_dev, PM_DEVICE_ACTION_SUSPEND);
    pm_device_action_run(flash_dev, PM_DEVICE_ACTION_SUSPEND);
#if CONFIG_NORDIC_QSPI_NOR_XIP
    xip_enabled = 0;
    nrf_qspi_nor_xip_enable(flash_dev, false);
#endif

    dk_set_led(RUN_STATUS_LED, (++blink_status) % 2);
    k_sleep(K_SECONDS(10));

    pm_device_action_run(uart_dev, PM_DEVICE_ACTION_RESUME);
    pm_device_action_run(flash_dev, PM_DEVICE_ACTION_RESUME);
    xip_enabled = 1;
  }
}
